// softrender by tmtlib

unit softrender;

interface

uses uqpixels, windows,graphics,math,dialogs,sysutils,jpeg;

type TRGB=packed record
           case integer of
            0:(arr:array [0..2] of byte);
            1:(r,g,b:byte);
            end;

type TBACK=packed record
                   poly_id:integer;
                   tex_id:integer;
                   u,v,z:single;
                  end;

type TBackRGB=array of array of TRGB;
type TBackInteger=array of array of integer;
type TBackSingle=array of array of single;
type TBackPlane=array of array of TBACK;

const max_plane=15;



type TScanPoint=record
                 _x:array[0..1] of integer;
                 _z,_u,_v:array[0..1] of single;


                 poly_id:integer;
                 tex_id:integer;
                end;

type TScanLine=array of TScanPoint;


type TBackBuffer=record
                  Bitmap:TBitmap;
                  QP:TQuickPixels;
                  Poly_now:integer; // current polygon number
                  Tex_now:integer;   // current texture
                  Scanlines:array of TScanLine;
                  Scancounts:array of integer;

                  DepthLines:TBackSingle;
                  PolyLines:TBackInteger;

                 end;


var BackBuffer:TBackBuffer;

var Textures:array of TBitMap;
var TextureQPS:array of TQuickPixels;

var xcount:integer;




procedure BufferSetDim(x,y:integer);
procedure BufferLock;
procedure BufferUnLock;
procedure SetPixel(x,y:integer; color:TColor);


procedure Polygon(polyId:integer);

procedure Rasterize;
function loadTex(fname:string):integer;

type TMatrix = packed record
                case integer of
                 0:(m:array [0..15] of Single);
                 1:(mi:array[0..3,0..3] of Single);
                end;




type TIndex = packed record
               case integer of
                0:(v1,v2,v3,n1,n2,n3,texId:Integer);
                1:(arr:array[0..6] of Integer);
                2:(vert:array[0..2] of Integer; norm:array[0..2] of integer; texture:integer);
               end;

type TTexCoord = packed record
                  u,v:single;
                 end;

type TTexCoords = packed record
                   case integer of
                    0:(u1,v1,u2,v2,u3,v3:single);
                    1:(uv:array[0..2] of TTexCoord);
                   end;

type TPoly = record
              Index     : TIndex;
              TexCoords : TTexCoords;
              end;

type TPolys = array of TPoly;


type TVect=packed record
              case integer of
               0:(arr:array[0..2] of Single);
               1:(x,y,z:Single);
              end;

type TRasterVect=packed record
                  case integer of
                  0:(arr:array[0..5] of Single);
                  1:(vert:TVect; uv:TTexCoord);
                  2:(x,y,z,u,v:single);
                 end;

type TVect4=packed record
              case integer of
               0:(arr:array[0..3] of Single);
               1:(x,y,z,w:Single);
              end;

type TVects=array of TVect;

var Polys:TPolys;
var Verts:TVects;
var Norms:TVects;


var matProj,
    matMod:TMatrix;
    viewPort:TVect4;

var camera:TVect;
var Tang:TVect;
var Norm:TVect;
var BiNorm:TVect;

var _p:array[0..2] of TRasterVect;
procedure RLine(polyId:integer; id1,id2:integer);


procedure _gluPerspective(fovy, aspect, zNear, zFar: Single);
function _gluProject(vertex:TVect):TVect;
procedure _gluMakeIdentity(var m:TMatrix);

procedure dxuLookAt(eyex, eyey, eyez,
                    centerx, centery, centerz,
                    upx, upy, upz: single);


function dxTranslatef (x,y,z:single):TMatrix;

function dxRotatef( angle,  x,  y, z:single):TMatrix;
procedure Mult(mat:TMatrix);
procedure Poly(p:array of TVect; tId:integer;texcoords:TTexCoords);
function Vector(p1,p2,p3:single):TVect;
function UV(u1,v1,u2,v2,u3,v3:single):TTexCoords;

procedure Line(v1,v2:TVect);

procedure SetDepth(x,y:integer; depth:single);
function GetDepth(x,y:integer):single;


implementation

function UV(u1,v1,u2,v2,u3,v3:single):TTexCoords;
begin
 result.u1:=u1;
 result.v1:=v1;
 result.u2:=u2;
 result.v2:=v2;
 result.u3:=u3;
 result.v3:=v3;
end;

function AddVert(vert:TVect):integer;
var i:integer;
begin
 i:=length(verts);
 setlength(verts,i+1);
 verts[i]:=vert;
 result:=i;
end;

function AddNorm(norm:TVect):integer;
var i:integer;
begin
 i:=length(norms);
 setlength(norms,i+1);
 norms[i]:=norm;
 result:=i;
end;

procedure AddPoly(index:TIndex; texcoords:TTexCoords);
var i:integer;
begin
 i:=length(Polys);
 setlength(Polys,i+1);
 Polys[i].Index:=index;
 Polys[i].TexCoords:=texcoords;
end;

procedure Poly(p:array of TVect; tId:integer;texcoords:TTexCoords);
var Index:TIndex;
var i:integer;
begin
  for i := 0 to 2 do
   Index.arr[i]:=AddVert(p[i]); // v1v2v3
  for i := 3 to 5 do
   Index.arr[i]:=AddNorm(p[i]); // n1n2n3
   Index.arr[6]:=tId;

   AddPoly(index,texcoords);

end;

function Sum(v1,v2:TVect):Tvect;
var i:integer;
begin
for i := 0 to 2 do
 result.arr[i]:=v1.arr[i]+v2.arr[i];
end;
function Vector(p1,p2,p3:single):TVect;
begin
 result.x:=p1;
 result.y:=p2;
 result.z:=p3;
end;


function dot(v1,v2:Tvect):single;
begin
result:=v1.x*v2.x+
        v1.y*v2.y+
        v1.z*v2.z;
end;




function dxRotatef( angle,  x,  y, z:single):TMatrix;
var mag, s, c:single;
    xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c:single;
begin

   s:= sin( angle * Pi/180 );
   c:= cos( angle * Pi/180 );

   mag:= SQRT( x*x + y*y + z*z );

   if (mag = 0.0) then
    begin
     _gluMakeIdentity(result);
     exit;
    end;

   x:=x/ mag;
   y:=y/ mag;
   z:=z/ mag;

   xx:= x * x;
   yy:= y * y;
   zz:= z * z;
   xy:= x * y;
   yz:= y * z;
   zx:= z * x;
   xs:= x * s;
   ys:= y * s;
   zs:= z * s;
   one_c:= 1.0 - c;

   result.mi[0,0]:= (one_c * xx) + c;
   result.mi[0,1]:= (one_c * xy) - zs;
   result.mi[0,2]:= (one_c * zx) + ys;
   result.mi[0,3]:= 0.0;

   result.mi[1,0]:= (one_c * xy) + zs;
   result.mi[1,1]:= (one_c * yy) + c;
   result.mi[1,2]:= (one_c * yz) - xs;
   result.mi[1,3]:= 0.0;

   result.mi[2,0]:= (one_c * zx) - ys;
   result.mi[2,1]:= (one_c * yz) + xs;
   result.mi[2,2]:= (one_c * zz) + c;
   result.mi[2,3]:= 0.0;

   result.mi[3,0]:= 0.0;
   result.mi[3,1]:= 0.0;
   result.mi[3,2]:= 0.0;
   result.mi[3,3]:= 1.0;

Mult(result);
end;



function dxTranslatef (x,y,z:single):TMatrix;
begin
  _gluMakeIdentity(result);
  result.mi[3,0]:=x;
  result.mi[3,1]:=y;
  result.mi[3,2]:=z;
 mult(result);
end;


procedure Mult(mat:TMatrix);
var row,clm: integer;
var res:TMatrix;
begin

 for clm := 0 to 3 do
  for row := 0 to 3 do
    res.mI[clm,row] := MatMod.mI[0,row] * mat.mi[clm,0] +
                       matMod.mI[1,row] * mat.mi[clm,1] +
                       matMod.mI[2,row] * mat.mi[clm,2] +
                       matMod.mI[3,row] * mat.mi[clm,3];


 for clm := 0 to 15 do
   matmod.m[clm]:=res.m[clm];
end;

function Trans(mat:TMatrix):TMatrix;
var row,clm:integer;
begin
 for clm := 0 to 3 do
  for row := 0 to 3 do
   result.mi[row,clm]:=mat.mi[clm,row];

end;


procedure normalize(var v:TVect);
var r:single;
begin
    r:= sqrt( v.x*v.x + v.y*v.y + v.z*v.z );
    if (r = 0.0) then exit;
    v.x:=v.x / r;
    v.y:=v.y / r;
    v.z:=v.z / r;
end;

procedure cross(v1:TVect; v2:TVect;var xresult:TVect);
begin
    xresult.x:= v1.y*v2.z - v1.z*v2.y;
    xresult.y:= v1.z*v2.x - v1.x*v2.z;
    xresult.z:= v1.x*v2.y - v1.y*v2.x;
end;

procedure __gluMakeIdentityf(var m:TMatrix);
begin
    m.m[0+4*0]:= 1; m.m[0+4*1]:= 0; m.m[0+4*2]:= 0; m.m[0+4*3]:= 0;
    m.m[1+4*0]:= 0; m.m[1+4*1]:= 1; m.m[1+4*2]:= 0; m.m[1+4*3]:= 0;
    m.m[2+4*0]:= 0; m.m[2+4*1]:= 0; m.m[2+4*2]:= 1; m.m[2+4*3]:= 0;
    m.m[3+4*0]:= 0; m.m[3+4*1]:= 0; m.m[3+4*2]:= 0; m.m[3+4*3]:= 1;
end;


procedure dxuLookAt(eyex, eyey, eyez,
                    centerx, centery, centerz,
                    upx, upy, upz: single);
var xforward, side, up:TVect;
var m:TMatrix;
begin
    xforward.x:= centerx - eyex;
    xforward.y:= centery - eyey;
    xforward.z:= centerz - eyez;

    up.x:= upx;
    up.y:= upy;
    up.z:= upz;

    normalize(xforward);

    //* Side = forward x up */
    cross(xforward, up, side);
    normalize(side);

    //* Recompute up as: up = side x forward */
    cross(side, xforward, up);

    __gluMakeIdentityf(m);
    m.mi[0][0]:= side.x;
    m.mi[1][0]:= side.y;
    m.mi[2][0]:= side.z;

    m.mi[0][1]:= up.x;
    m.mi[1][1]:= up.y;
    m.mi[2][1]:= up.z;

    m.mi[0][2]:= -xforward.x;
    m.mi[1][2]:= -xforward.y;
    m.mi[2][2]:= -xforward.z;


    mult(m);
//    glMultMatrixf(&m[0][0]);
    dxTranslatef(-eyex, -eyey, -eyez);


end;



procedure _gluMakeIdentity(var m:TMatrix);
begin
with m do
 begin
    m[0+4*0]:= 1; m[0+4*1]:= 0; m[0+4*2]:= 0; m[0+4*3]:= 0;
    m[1+4*0]:= 0; m[1+4*1]:= 1; m[1+4*2]:= 0; m[1+4*3]:= 0;
    m[2+4*0]:= 0; m[2+4*1]:= 0; m[2+4*2]:= 1; m[2+4*3]:= 0;
    m[3+4*0]:= 0; m[3+4*1]:= 0; m[3+4*2]:= 0; m[3+4*3]:= 1;
 end;
end;


procedure _gluPerspective(fovy, aspect, zNear, zFar: Single);
var sine, cotangent, deltaZ:single;
    radians:single;
begin
  radians:=fovy / 2 * Pi / 180;
  deltaZ:= zFar - zNear;
  sine:= sin(radians);
  if ((deltaZ = 0) or (sine = 0) or (aspect = 0)) then exit;
  cotangent:= COS(radians) / sine;

    _gluMakeIdentity(matProj);
    with matProj do
     begin
       mi[0][0]:= cotangent / aspect;
       mi[1][1]:= cotangent;
       mi[2][2]:= -(zFar + zNear) / deltaZ;
       mi[2][3]:= -1;
       mi[3][2]:= -2 * zNear * zFar / deltaZ;
       mi[3][3]:= 0;
     end;
//matProj:=glMultMatrix(matProj,m);
end;


function _gluMultMatrixVec(matrix:TMatrix; _in:TVect4):TVect4;
var i:integer;
begin

for I := 0 to 3 do
 result.arr[i]:=
	    _in.arr[0] * matrix.m[0*4+i] +
	    _in.arr[1] * matrix.m[1*4+i] +
	    _in.arr[2] * matrix.m[2*4+i] +
	    _in.arr[3] * matrix.m[3*4+i];
end;


function _gluProject(vertex:TVect):TVect;
var _in,_out:TVect4;
begin


    _in.x:=vertex.x;
    _in.y:=vertex.y;
    _in.z:=vertex.z;
    _in.w:=1.0;
    _out:=_gluMultMatrixVec(matMod, _in);
    _in:=_gluMultMatrixVec(matProj, _out);


    if (_in.w = 0.0) then exit;
    _in.x:=_in.x / _in.w;
    _in.y:=_in.y / _in.w;
    _in.z:=_in.z / _in.w;

    //* Map x, y and z to range 0-1 */
    _in.x:= _in.x * 0.5 + 0.5;
    _in.y:= _in.y * 0.5 + 0.5;
    _in.z:= _in.z * 0.5 + 0.5;

    //* Map x,y to viewport */
    _in.x:= _in.x * viewport.arr[2] + viewport.arr[0];
    _in.y:= _in.y * viewport.arr[3] + viewport.arr[1];


    result.arr:=vertex.arr;
    result.x:=_in.x;
    result.y:=_in.y;
//    result.z:=_in.z;

    //hack
    result.z:=_in.w;



end;










function loadTex(fname:string):integer;
var i:integer;
var
  Picture: TPicture;
begin

Picture := TPicture.Create;
Picture.LoadFromFile(fname);

i:=length(Textures);
setlength(Textures,i+1);
Textures[i]:=TBitmap.Create;

Textures[i].Width:=Picture.Width;
Textures[i].Height:=Picture.Height;
Textures[i].PixelFormat:=pf32bit;
Textures[i].Canvas.Draw(0,0,Picture.Graphic);
Picture.Free;


setlength(TextureQPS,i+1);
TextureQPS[i]:=TQuickPixels.Create;
TextureQPS[i].Attach(Textures[i]);
result:=i;

end;

procedure BufferSetDim(x,y:integer);
var i:integer;
begin
  BackBuffer.Bitmap:=TBitmap.Create;
  BackBuffer.Bitmap.Width:=x;
  BackBuffer.Bitmap.Height:=y;
  BackBuffer.Bitmap.PixelFormat:=pf32bit;


  setlength(BackBuffer.Scanlines,y);
  setlength(BackBuffer.Scancounts,y);  
  setlength(BackBuffer.Depthlines,y,x);
  setlength(BackBuffer.PolyLines,y,x);

end;


procedure BufferLock;
var x,y:integer;
begin
 with BackBuffer do
  begin
   Poly_now:=0;
   QP:=TQuickPixels.Create;
   QP.Attach(Bitmap);
  end;



 for y := 0 to length(BackBuffer.Scanlines) - 1 do
  begin
   setlength(BackBuffer.Scanlines[y],0);
   BackBuffer.Scancounts[y]:=0;
  end;

 for y := 0 to length(BackBuffer.Depthlines) - 1 do
  for x := 0 to length(BackBuffer.Depthlines[0]) - 1 do
  begin
   BackBuffer.Depthlines[y,x]:=0;
   BackBuffer.PolyLines[y,x]:=-1;
  end;



end;

procedure BufferUnLock;
begin
  BackBuffer.QP.Free;
end;

procedure SetDepth(x,y:integer; depth:single);
begin
 if (x>=0) and (y>=0) and (x<=BackBuffer.Bitmap.Width-1) and
 (y<=BackBuffer.Bitmap.Height-1) then
 BackBuffer.Depthlines[y][x]:=depth;
end;

function GetDepth(x,y:integer):single;
begin
 if (x>=0) and (y>=0) and (x<=BackBuffer.Bitmap.Width-1) and
 (y<=BackBuffer.Bitmap.Height-1) then
 result:=BackBuffer.Depthlines[y][x];
end;

procedure SetPoly(x,y:integer; depth:integer);
begin
 if (x>=0) and (y>=0) and (x<=BackBuffer.Bitmap.Width-1) and
 (y<=BackBuffer.Bitmap.Height-1) then
 BackBuffer.PolyLines[y][x]:=depth;
end;

function GetPoly(x,y:integer):integer;
begin
 if (x>=0) and (y>=0) and (x<=BackBuffer.Bitmap.Width-1) and
 (y<=BackBuffer.Bitmap.Height-1) then
 result:=BackBuffer.PolyLines[y][x];
end;

procedure SetPixel(x,y:integer; Color:TColor);
begin
 if (x>=0) and (y>=0) and (x<=BackBuffer.Bitmap.Width-1) and
 (y<=BackBuffer.Bitmap.Height-1) then

  BackBuffer.QP.SetPixel(x,BackBuffer.Bitmap.Height-1-y,Color);


end;



type TPolyx=record
             left_x:integer;
             right_x:integer;

             u1,v1,z1:single;
             u2,v2,z2:single;

             poly_idx:integer;
             tex_idx:integer;
            end;

var polyx:array of TPolyx;

procedure Rasterize;
var x,y,i,m,k:integer;
var found:boolean;
var found_idx:integer;
var u,v,z:single;
var du,dv,dz:single;
var dx,tex:integer;
var Color:TColor;
var tu,tv:integer;
var tid:integer;

//var point1,point2:TScanPoint;

dk:integer;

var u0,v0,z0:single;

begin

xcount:=0;

  for y := length(BackBuffer.Scanlines)-1 downto 0 do
   for i := 0 to length(BackBuffer.Scanlines[y]) - 1 do
   with BackBUffer.Scanlines[y][i] do
    begin
    if (_x[0]<0) and
       (_x[1]<0) then poly_id:=-1;

    if (_x[0]>BackBuffer.Bitmap.Width-1) and
       (_x[1]>BackBuffer.Bitmap.Width-1) then poly_id:=-1;
    end;


       {
  for y := 0 to length(BackBuffer.Scanlines)-1 do
   for k:=0 to BackBuffer.Bitmap.Width- 1 do
//   BackBuffer.QP.SetPixel(k,y,rgb(0,255,0));   
    SetPixel(k,y,rgb(0,255,0));   
//             SetDepth(k,y,0);
        }


//if false then

  for y := 0 to length(BackBuffer.Scanlines)-1 do
   for i := 0 to length(BackBuffer.Scanlines[y])- 1 do
    if BackBUffer.Scanlines[y][i].poly_id>=0 then
   with BackBuffer.Scanlines[y][i] do    
   begin
              



    if (_z[0]<1/5) and (_z[1]<1/5) then
    if not ((_x[0]<0) and (_x[1]<0)) then
    if not ((_x[0]>BackBuffer.Bitmap.Width-1) and (_x[1]>BackBuffer.Bitmap.Width-1)) then

     if (_z[0]>GetDepth(_x[0],y)-1/100) or
        (_z[1]>GetDepth(_x[1],y)-1/100) then
    begin

        tid:=tex_id;

        u:=_u[0];
        v:=_v[0];
        z:=_z[0];

        dx:=(_x[1]-_x[0]);
        dk:=sign(dx);
        dx:=abs(dx);


       if dx=0 then // dirty
       begin
         du:=0;
         dv:=0;
         dz:=0;
       end else
       begin

        du:=(_u[1]-_u[0])/dx;
        dv:=(_v[1]-_v[0])/dx;
        dz:=(_z[1]-_z[0])/dx;
       end;


       
        k:=_x[0];


        repeat


           if z>GetDepth(k,y) then
            begin



             SetDepth(k,y,z);
            SetPoly(k,y,poly_id);

             inc(xcount);


           if z>0 then
            begin
             z0:=1/z;
             u0:=u*z0;
             v0:=v*z0;
            end;


           tu:=round(u0*(Textures[tid].Width-1));
           tv:=round(v0*(Textures[tid].Height-1));

          if (tu>=0) and (tv>=0) and
           (tu<=Textures[tid].Width-1) and
           (tv<=Textures[tid].Height-1) then
           Color:=TextureQPS[tid].GetPixel(tu,tv) else
                      
           Color:=rgb(0,128,0);



           SetPixel(k,y,Color);

            end;      

           u:=u+du;
           v:=v+dv;
           z:=z+dz;       
                                  
           k:=k+dk;
                                  
          until k=_x[1]+dk;

                                  


    end;                          
               



   end;




end;



procedure Polygon(polyId:integer);
var i,k,m:integer;
begin
  BackBUffer.tex_now:=polys[polyId].Index.texId;


  for i := 0 to 2 do
   begin
    k:=polys[polyId].Index.vert[i];
    _p[i].vert:=_gluProject(verts[k]);
    _p[i].uv:=polys[polyId].TexCoords.uv[i];
   end;


 if _p[0].z>1 then
 if _p[1].z>1 then
 if _p[2].z>1 then

 begin

   for i := 0 to 2 do
    if _p[i].z>0 then
     begin
      _p[i].z:=1/_p[i].z;
      _p[i].u:=_p[i].u*_p[i].z;
      _p[i].v:=_p[i].v*_p[i].z;
     end else _p[i].z:=0;


  RLine(polyId,0,1);
  RLine(polyId,1,2);
  RLine(polyId,2,0);

  inc(BackBuffer.Poly_now);

  for i := 0 to length(BackBuffer.Scanlines) - 1 do
   BackBuffer.Scancounts[i]:=0;

 end;

                             {
  for i := 0 to 2 do
   begin
    k:=polys[polyId].Index.vert[i];
    m:=polys[polyId].Index.norm[i];

    Line(verts[k],sum(verts[k],norms[m]));
   end;
                              }


end;

var zzz:integer;
procedure AddScanPoint(x,y:integer;z,u,v:single;tid:integer);
var i,id:integer;
var enab:boolean;
begin

 if (y<0) or (y>length(BackBuffer.Scanlines)-1) then exit;
    SetPixel(x,y,255);

 i:=length(BackBuffer.Scanlines[y])-1;
 id:= BackBuffer.Scancounts[y];

 if id=0 then
  begin
   i:=i+1;
   setlength(BackBuffer.Scanlines[y],i+1);
   BackBuffer.Scanlines[y][i].poly_id:=BackBuffer.poly_now;
  end;


 Enab:=true;
 if id>1 then
  begin
   id:=1;
   if not((BackBuffer.Scanlines[y][i]._x[0]<>x) and
          (BackBuffer.Scanlines[y][i]._x[1]<>x)) then enab:=false;
  end;

 if enab then
  begin

 BackBuffer.Scanlines[y][i]._x[id]:=x;
 BackBuffer.Scanlines[y][i]._z[id]:=z;
 BackBuffer.Scanlines[y][i]._u[id]:=u;
 BackBuffer.Scanlines[y][i]._v[id]:=v;
 BackBuffer.Scanlines[y][i].tex_id:=tid;
  end;




 inc(id);

 BackBuffer.Scancounts[y]:=id;




end;

procedure AddHackPoint(x,y:integer);
begin
 if (y<0) or (y>length(BackBuffer.Scanlines)-1) then exit;
            SetPixel(x,y,255);
end;



procedure Line(v1,v2:TVect);
var x,y,dx,dy:single;
var i,di:integer;
//var u,v,z,du,dv,dz:single;
oldx,oldy:single;
var ye:integer;
begin

_p[0].vert:=_gluProject(v1);
_p[1].vert:=_gluProject(v2);

dx:=ceil(_p[1].x)-ceil(_p[0].x);
dy:=ceil(_p[1].y)-ceil(_p[0].y);



if dy=0 then
begin
 AddHackPoint(ceil(_p[0].x),ceil(_p[0].y));
 AddHackPoint(ceil(_p[1].x),ceil(_p[1].y));
end
else
  begin
   dx:=dx/abs(dy);
   di:=sign(dy);
   ye:=ceil(_p[1].y);
   x:=_p[0].x;



    i:=ceil(_p[0].y);
    while i<>ye do
     begin

      AddHackPoint(ceil(x),i);

      x:=x+dx;
      i:=i+di;



      end;
//                       f
  //    AddHackPoint(ceil(x),i);


  end;


end;

procedure RLine(polyId:integer; id1,id2:integer);
var x,y,dx,dy:single;
var i,di:integer;
var u,v,z,du,dv,dz:single;
oldx,oldy:single;
var ye:integer;

begin




dx:=_p[id2].x-_p[id1].x;
dy:=ceil(_p[id2].y)-ceil(_p[id1].y);


du:=_p[id2].u-_p[id1].u;
dv:=_p[id2].v-_p[id1].v;
dz:=_p[id2].z-_p[id1].z;

u:=_p[id1].u;
v:=_p[id1].v;
z:=_p[id1].z;



if dy=0 then
begin
 AddScanPoint(ceil(_p[id1].x),ceil(_p[id1].y),_p[id1].z,_p[id1].u,_p[id1].v,BackBuffer.Tex_now);
 AddScanPoint(ceil(_p[id2].x),ceil(_p[id2].y),_p[id2].z,_p[id2].u,_p[id2].v,BackBuffer.Tex_now);
end
else
  begin
   dx:=dx/abs(dy);
   du:=du/abs(dy);
   dv:=dv/abs(dy);
   dz:=dz/abs(dy);
   di:=sign(dy);
   ye:=ceil(_p[id2].y)+di;
   x:=_p[id1].x;




    i:=ceil(_p[id1].y);
    while i<>ye do
     begin
        zzz:=id1;
      AddScanPoint(ceil(x),i,z,u,v,BackBuffer.Tex_now);

      x:=x+dx;
      u:=u+du;
      v:=v+dv;
      z:=z+dz;
      i:=i+di;

      end;

//      AddScanPoint(ceil(_p[id2].x),i,z,u,v,BackBuffer.Tex_now);


  end;





end;


end.
