// OpenGL -> Direct3D wrapper by Georgy Moshkin (tmtlib@narod.ru)
// v0.1
// based on MESA3D, labo.xo-ox.net, xproger, ...

unit tinyd3d;

interface

uses math;


type TMatrix = packed record
                case integer of
                 0:(m:array [0..15] of Single);
                 1:(mi:array[0..3,0..3] of Single);
                end;



function dxRotatef( angle,  x,  y, z:single):TMatrix;
function dxTranslatef (x,y,z:single):TMatrix;



function PerspectiveFovLH(fovy, aspect, zNear, zFar: Single): TMatrix;

procedure dxuLookAt(eyex, eyey, eyez,
                    centerx, centery, centerz,
                    upx, upy, upz: single);


procedure MatrixIdentity(var mat: TMatrix);
procedure MatrixMultiply(var mat:TMatrix;  mat1,mat2:TMatrix);
procedure Transpose(var mat:TMatrix);


const DX_PROJECTION=0;
const DX_MODELVIEW=1;  //DX_D3DTS_WORLD
const DX_D3DTS_VIEW=2;

var Matrix:array[0..2] of TMatrix;
var MatrixMode:integer;



procedure dxMatrixMode(mode:integer);
procedure dxLoadIdentity;
procedure dxuPerspective(fovy, aspect, zNear, zFar: Single);


var PushMatrix:array of TMatrix;

procedure dxPushMatrix;
procedure dxPopMatrix;


implementation

procedure dxPushMatrix;
var i:integer;
begin
 i:=length(PushMatrix);
 setlength(PushMatrix,i+1);
 PushMatrix[i]:=Matrix[DX_MODELVIEW];

end;

procedure dxPopMatrix;
var i:integer;
begin
 i:=length(PushMatrix)-1;
 Matrix[DX_MODELVIEW]:=PushMatrix[i];
 setlength(PushMatrix,i);

end;

{
procedure dxEnd;
begin
 device9.BeginScene;
 device9.SetFVF(D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_DIFFUSE or D3DFVF_TEX1);
 Transpose(Matrix[DX_MODELVIEW]); //gl->dx
 device9.SetTransform(D3DTS_WORLD,Matrix[DX_MODELVIEW]);
 device9.DrawPrimitiveUP(D3DPT_TRIANGLELIST, (xVertInd+1) div 3, xVertex, sizeof(TDXVertex));
 Transpose(Matrix[DX_MODELVIEW]); //dx->gl
 device9.EndScene;
 end;
 }

procedure dxMatrixMode(mode:integer);
begin
 MatrixMode:=mode;
end;

procedure dxLoadIdentity;
begin
MatrixIdentity(Matrix[MatrixMode])

{;
  case MatrixMode of
    DX_PROJECTION: MatrixIdentity(Matrix[MatrixMode]);
    DX_MODELVIEW:
    begin
     MatrixIdentity(Matrix[MatrixMode]);  //DX_D3DTS_WORLD
//     MatrixIdentity(Matrix[DX_D3DTS_VIEW]);

    end;
    end;}
end;

procedure dxuPerspective(fovy, aspect, zNear, zFar: Single);
begin
 MatrixMode:= DX_PROJECTION;
 Matrix[MatrixMode]:=PerspectiveFovLH(fovy, aspect, zNear, zFar);
 Transpose(Matrix[MatrixMode]);
// device9.SetTransform(D3DTS_PROJECTION, Matrix[MatrixMode]);

end;


function PerspectiveFovLH(fovy, aspect, zNear, zFar: Single): TMatrix;
var xmin, xmax, ymin, ymax:single;
    x, y, a, b, c, d:single;
    m:array[0..16-1] of single;
    left, right,
    bottom, top,
    nearval, farval:single;

begin
   ymax:= zNear * tan(fovy * PI / 360.0);
   ymin:= -ymax;
   xmin:= ymin * aspect;
   xmax:= ymax * aspect;


   left:=xmin;
   right:=xmax;
   bottom:=ymin;
   top:=ymax;
   nearval:=zNear;
   farval:=zFar;


   x:= (2.0 * nearval) / (right - left);
   y:= (2.0 * nearval) / (top - bottom);
   a:= (right + left) / (right - left);
   b:= (top + bottom) / (top - bottom);
   c:= -(farval + nearval) / ( farval - nearval);
   d:= -(2.0 * farval * nearval) / (farval - nearval);

   result.Mi[0,0]:=x;    result.Mi[0,1]:= 0.0;  result.Mi[0,2]:= a;     result.Mi[0,3]:= 0.0;
   result.Mi[1,0]:=0.0;  result.Mi[1,1]:= y;    result.Mi[1,2]:= b;     result.Mi[1,3]:= 0.0;
   result.Mi[2,0]:=0.0;  result.Mi[2,1]:= 0.0;  result.Mi[2,2]:= c;     result.Mi[2,3]:= d;
   result.Mi[3,0]:=0.0;  result.Mi[3,1]:= 0.0;  result.Mi[3,2]:= -1.0;  result.Mi[3,3]:= 0.0;

end;

procedure dxuLookAt(eyex, eyey, eyez,
                    centerx, centery, centerz,
                    upx, upy, upz: single);
var x, y, z:array[0..2] of single;
mag:single;
mmm:TMatrix;
begin

   z[0]:= eyex - centerx;
   z[1]:= eyey - centery;
   z[2]:= eyez - centerz;
   mag:= sqrt(z[0] * z[0] + z[1] * z[1] + z[2] * z[2]);
   if (mag>0) then begin		 //	/* mpichler, 19950515 */
      z[0]:=z[0]/ mag;
      z[1]:=z[1]/ mag;
      z[2]:=z[2]/ mag;
    end;


   y[0]:= upx;
   y[1]:= upy;
   y[2]:= upz;

   ///* X vector = Y cross Z */
   x[0]:= y[1] * z[2] - y[2] * z[1];
   x[1]:= -y[0] * z[2] + y[2] * z[0];
   x[2]:= y[0] * z[1] - y[1] * z[0];

   ///* Recompute Y = Z cross X */
   y[0]:= z[1] * x[2] - z[2] * x[1];
   y[1]:= -z[0] * x[2] + z[2] * x[0];
   y[2]:= z[0] * x[1] - z[1] * x[0];

   {
   /* mpichler, 19950515 */
   /* cross product gives area of parallelogram, which is < 1.0 for
    * non-perpendicular unit-length vectors; so normalize x, y here
    */
    }
   mag:= sqrt(x[0] * x[0] + x[1] * x[1] + x[2] * x[2]);
   if (mag>0) then
    begin
      x[0]:=x[0] / mag;
      x[1]:=x[1] / mag;
      x[2]:=x[2] / mag;
    end;

   mag:= sqrt(y[0] * y[0] + y[1] * y[1] + y[2] * y[2]);
   if (mag>0) then
    begin
      y[0]:=y[0] / mag;
      y[1]:=y[1] / mag;
      y[2]:=y[2] / mag;
    end;

   Matrix[DX_D3DTS_VIEW].Mi[0,0]:= x[0];
   Matrix[DX_D3DTS_VIEW].Mi[0,1]:= x[1];
   Matrix[DX_D3DTS_VIEW].Mi[0,2]:= x[2];
   Matrix[DX_D3DTS_VIEW].Mi[0,3]:= 0.0;
   Matrix[DX_D3DTS_VIEW].Mi[1,0]:= y[0];
   Matrix[DX_D3DTS_VIEW].Mi[1,1]:= y[1];
   Matrix[DX_D3DTS_VIEW].Mi[1,2]:= y[2];
   Matrix[DX_D3DTS_VIEW].Mi[1,3]:= 0.0;
   Matrix[DX_D3DTS_VIEW].Mi[2,0]:= z[0];
   Matrix[DX_D3DTS_VIEW].Mi[2,1]:= z[1];
   Matrix[DX_D3DTS_VIEW].Mi[2,2]:= z[2];
   Matrix[DX_D3DTS_VIEW].Mi[2,3]:= 0.0;
   Matrix[DX_D3DTS_VIEW].Mi[3,0]:= 0.0;
   Matrix[DX_D3DTS_VIEW].Mi[3,1]:= 0.0;
   Matrix[DX_D3DTS_VIEW].Mi[3,2]:= 0.0;
   Matrix[DX_D3DTS_VIEW].Mi[3,3]:= 1.0;



  mmm:= dxTranslatef (-eyex, -eyey, -eyez);
  MatrixMultiply(Matrix[DX_D3DTS_VIEW],mmm,Matrix[DX_D3DTS_VIEW]);

//  Transpose(Matrix[DX_D3DTS_VIEW]);
//  device9.SetTransform(D3DTS_VIEW, Matrix[DX_D3DTS_VIEW]);
end;

function dxRotatef( angle,  x,  y, z:single):TMatrix;
var mag, s, c:single;
    xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c:single;
begin

   s:= sin( angle * Pi/180 );
   c:= cos( angle * Pi/180 );

   mag:= SQRT( x*x + y*y + z*z );

   if (mag = 0.0) then
    begin
     MatrixIdentity(result);
     exit;
    end;

   x:=x/ mag;
   y:=y/ mag;
   z:=z/ mag;

   xx:= x * x;
   yy:= y * y;
   zz:= z * z;
   xy:= x * y;
   yz:= y * z;
   zx:= z * x;
   xs:= x * s;
   ys:= y * s;
   zs:= z * s;
   one_c:= 1.0 - c;

   result.Mi[0,0]:= (one_c * xx) + c;
   result.Mi[0,1]:= (one_c * xy) - zs;
   result.Mi[0,2]:= (one_c * zx) + ys;
   result.Mi[0,3]:= 0.0;

   result.Mi[1,0]:= (one_c * xy) + zs;
   result.Mi[1,1]:= (one_c * yy) + c;
   result.Mi[1,2]:= (one_c * yz) - xs;
   result.Mi[1,3]:= 0.0;

   result.Mi[2,0]:= (one_c * zx) - ys;
   result.Mi[2,1]:= (one_c * yz) + xs;
   result.Mi[2,2]:= (one_c * zz) + c;
   result.Mi[2,3]:= 0.0;

   result.Mi[3,0]:= 0.0;
   result.Mi[3,1]:= 0.0;
   result.Mi[3,2]:= 0.0;
   result.Mi[3,3]:= 1.0;

   MatrixMultiply(Matrix[MatrixMode],result,Matrix[MatrixMode]);
end;





procedure MatrixIdentity(var mat: TMatrix);
var i:integer;
begin
 FillChar(mat, SizeOf(mat), 0);
 for i := 0 to 3 do mat.mi[i,i]:=1;
end;



function dxTranslatef (x,y,z:single):TMatrix;
begin
  MatrixIdentity(result);
  result.mi[0,3]:=x;
  result.mi[1,3]:=y;
  result.mi[2,3]:=z;
  MatrixMultiply(Matrix[MatrixMode],result,Matrix[MatrixMode]);
end;

procedure Transpose(var mat:TMatrix);
var tmp:TMatrix;
Var row,clm: integer;
begin
tmp:=mat;
 for clm := 0 to 3 do
  for row := 0 to 3 do
   mat.mi[clm,row]:=tmp.mi[row,clm];
end;


procedure MatrixMultiply(var mat:TMatrix;  mat1,mat2:TMatrix);
var row,clm: integer;
begin

 for clm := 0 to 3 do
  for row := 0 to 3 do
    mat.Mi[clm,row] := mat1.Mi[0,row] * mat2.Mi[clm,0] +
                      mat1.Mi[1,row] * mat2.Mi[clm,1] +
                      mat1.Mi[2,row] * mat2.Mi[clm,2] +
                      mat1.Mi[3,row] * mat2.Mi[clm,3];
end;





end.
